# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from dataclasses import dataclass

# pylint:disable=too-few-public-methods


@dataclass
class Credentials:
    """
    Credentials for running circuits on an IBM backend.
    For more information on the fields, see:
    https://quantum-computing.ibm.com/lab/docs/iql/manage/account/ibmq

    Parameters
    ----------
    token : str
        An IBM account API token.
    project : str
        The IBM project.
    hub : str
        The IBM hub.
    group : str
        The IBM group.
    """

    token: str
    project: str
    hub: str
    group: str



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

import json
import time
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    Optional,
)

import gql
from qctrlclient import GraphQLClient

from .action_status import ActionStatus
from .registry import RemoteRegistry


class BaseRouter(ABC):
    """Routes a request to execute a workflow."""

    @abstractmethod
    def __call__(self, workflow: str, data: Optional[Dict[str, Any]] = None) -> Any:
        """Given a workflow name and the corresponding data (if any),
        executes the workflow and returns the raw result.

        Parameters
        ----------
        workflow : str
            Name of the workflow to be executed.
        data : Dict[str, Any], optional
            Any data required by the workflow for execution.
        """


class ApiRouter(BaseRouter):
    """Remotely executes the workflow using the `startCoreWorkflow`
    GraphQL mutation.

    Parameters
    ----------
    client : GraphQLClient
        The GraphQL client used to make the request to execute
        the workflow remotely.
    registry : Registry
        The registry that the workflows being executed are
        registered in.
    """

    _COMPLETED_STATES = (ActionStatus.SUCCESS.value, ActionStatus.FAILURE.value)

    def __init__(self, client: GraphQLClient, registry: RemoteRegistry):
        self._client = client
        self._registry = registry

    def __call__(self, workflow, data=None):

        query = gql.gql(
            """
            mutation ($input: StartCoreWorkflowInput!) {
                startCoreWorkflow(input: $input) {
                    action {
                        modelId
                        status
                        result
                        errors {
                            exception
                            traceback
                        }
                    }
                    errors {
                        message
                        fields
                    }
                }
            }
        """
        )

        input_ = {
            "registry": self._registry.value,
            "workflow": workflow,
            "data": json.dumps(data),
        }

        response = self._client.execute(query, {"input": input_})

        # pylint:disable=unsubscriptable-object

        action_id = response["startCoreWorkflow"]["action"]["modelId"]
        status = response["startCoreWorkflow"]["action"]["status"]
        result = response["startCoreWorkflow"]["action"]["result"]
        errors = response["startCoreWorkflow"]["action"]["errors"]

        result = self.poll_for_completion(action_id, status, result, errors)
        return result

    def poll_for_completion(
        self, action_id: str, status: str, result: Any, errors: Any
    ) -> Any:
        """Polls the API waiting for the action to be completed.
        When completed, the `result` is returned.
        """

        _query = gql.gql(
            """
            query($modelId: String!) {
                action(modelId: $modelId) {
                    action {
                        status
                        errors {
                            exception
                            traceback
                        }
                        result
                    }
                    errors {
                        message
                    }
                }
            }
        """
        )

        while status not in self._COMPLETED_STATES:
            time.sleep(2)  # FIXME use progressive polling
            response = self._client.execute(_query, {"modelId": action_id})
            status = response["action"]["action"]["status"]
            result = response["action"]["action"]["result"]
            errors = response["action"]["action"]["errors"]

        if status == ActionStatus.FAILURE.value:
            raise RuntimeError(errors)

        if result is not None:
            result = json.loads(result)

        return result


class LocalRouter(BaseRouter):
    """Executes workflows using a resolver provided by a local
    package which implements the workflows.

    Parameters
    ----------
    resolver : BaseResolver
        A resolver object for the registry which contains all
        required workflows.
    """

    def __init__(self, resolver: "BaseResolver"):
        self._resolver = resolver

    def __call__(self, workflow, data=None):
        data = data or {}
        task = self._resolver.get_workflow_task_from_signature(workflow, **data)
        func = self._resolver(task)
        result = func()
        return result



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from functools import wraps
from typing import Callable


def start_core_workflow(get_config: Callable, workflow: str):
    """Decorator for a function which will execute a workflow.
    The decorated function should return the data to be used
    during workflow execution. When being used in a client
    package, it is recommended to use a partial to provide
    a default value for `get_config` e.g.

    fire_opal_workflow = partial(
        start_core_workflow,
        get_fire_opal_config
    )

    @fire_opal_workflow("execute")
    def execute(...):

    Parameters
    ----------
    get_config : Callable
        Returns a `CoreClientSettings` instance. The configured
        router will be used to execute the workflow.
    workflow : str
        The registered name of the workflow to be executed.
    """

    def decorator(func: Callable):
        @wraps(func)
        def customized_decorator(*args, **kwargs):
            data = func(*args, **kwargs)
            config = get_config()
            router = config.get_router()
            result = router(workflow, data)
            return result

        return customized_decorator

    return decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fireopal', 'fireopal.core', 'fireopal.functions']

package_data = \
{'': ['*']}

install_requires = \
['qctrl-client>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'fire-opal',
    'version': '2.0.1',
    'description': 'Fire Opal Client',
    'long_description': "# Fire Opal Client\n\nThe Fire Opal Client package is a Python client for Q-CTRL's Fire Opal product.\n",
    'author': 'Q-CTRL',
    'author_email': 'support@q-ctrl.com',
    'maintainer': 'Q-CTRL',
    'maintainer_email': 'support@q-ctrl.com',
    'url': 'https://q-ctrl.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from dataclasses import dataclass
from typing import (
    Callable,
    Union,
)

from .router import BaseRouter


@dataclass
class CoreClientSettings:
    """Settings class for core clients.

    Parameters
    ----------
    router : Union[BaseRouter, Callable]
        The router to be used in the core client. Can be either
        an instance of `BaseRouter` or a callable which accepts
        no arguments and returns an instance of `BaseRouter`.
    """

    router: Union[BaseRouter, Callable]

    def update(self, **kwargs):
        """Updates settings fields."""

        for attr, value in kwargs.items():
            if not hasattr(self, attr):
                raise AttributeError(f"Invalid field: {attr}")

            setattr(self, attr, value)

    def get_router(self) -> BaseRouter:
        """Returns the configured router to be used by the client."""
        result = self.router

        if isinstance(result, BaseRouter):
            pass

        elif callable(result):
            result = result()

        else:
            raise ValueError(f"Invalid router: {result}")

        return result



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from qctrlclient import (
    CliAuth,
    GraphQLClient,
)
from qctrlclient.constants import DEFAULT_API_URL
from qctrlclient.globals import (
    get_default_cli_auth,
    global_value,
)

from fireopal.core import (
    ApiRouter,
    CoreClientSettings,
    LocalRouter,
    RemoteRegistry,
)


def get_default_router() -> ApiRouter:
    """Returns the default router that the Fire Opal
    client uses.
    """
    return ApiRouter(
        GraphQLClient(DEFAULT_API_URL, auth=get_default_cli_auth()),
        RemoteRegistry.FIRE_OPAL,
    )


@global_value("FIRE_OPAL_CONFIG")
def get_config() -> CoreClientSettings:
    """Returns the global Fire Opal settings."""
    return CoreClientSettings(router=get_default_router)


def configure(**kwargs):
    """Updates the global Fire Opal settings. See `CoreClientSettings`
    for details on which fields can be updated.
    """
    config = get_config()
    config.update(**kwargs)


def configure_api(api_url: str, oidc_url: str):
    """Convenience function to configure Fire Opal for API
    routing.

    Parameters
    ----------
    api_url : str
        URL of the GraphQL schema
    oidc_url : str
        Base URL of the OIDC provider e.g. Keycloak
    """
    configure(
        router=ApiRouter(
            GraphQLClient(api_url, CliAuth(oidc_url)),
            RemoteRegistry.FIRE_OPAL,
        )
    )


def configure_local(resolver: "BaseResolver"):
    """Convenience function to configure Fire Opal for local
    routing.

    Parameters
    ----------
    resolver : BaseResolver
        A local implementation of a workflow resolver which uses
        a registry that implements all of the available Fire Opal
        workflows
    """
    configure(router=LocalRouter(resolver))
